﻿Imports System.Runtime.InteropServices

Public Class DeCompress

#Region "API"
    <DllImport("Cabinet.dll", EntryPoint:="CreateCompressor")>
    Private Shared Function CreateCompressor(<[In]> Algorithm As COMPRESS_ALGORITHM,
                                             <[In]> AllocationRoutines As IntPtr,
                                             <Out> ByRef CompressorHandle As IntPtr) As <MarshalAs(UnmanagedType.Bool)> Boolean
    End Function

    <DllImport("Cabinet.dll", EntryPoint:="CreateDecompressor")>
    Private Shared Function CreateDecompressor(<[In]> Algorithm As COMPRESS_ALGORITHM,
                                               <[In]> AllocationRoutines As IntPtr,
                                               <Out> ByRef DecompressorHandle As IntPtr) As <MarshalAs(UnmanagedType.Bool)> Boolean
    End Function

    <DllImport("Cabinet.dll", EntryPoint:="CloseCompressor")>
    Private Shared Function CloseCompressor(<[In]> CompressorHandle As IntPtr) As <MarshalAs(UnmanagedType.Bool)> Boolean
    End Function

    <DllImport("Cabinet.dll", EntryPoint:="CloseDecompressor")>
    Private Shared Function CloseDecompressor(<[In]> DecompressorHandle As IntPtr) As <MarshalAs(UnmanagedType.Bool)> Boolean
    End Function

    <DllImport("Cabinet.dll", EntryPoint:="Compress")>
    Private Shared Function Compress(<[In]> CompressorHandle As IntPtr,
                                     <[In], MarshalAs(UnmanagedType.LPArray)> UncompressedData As Byte(),
                                     <[In]> UncompressedDataSize As Integer,
                                     <Out, MarshalAs(UnmanagedType.LPArray)> CompressedBuffer As Byte(),
                                     <[In]> CompressedBufferSize As Integer,
                                     <Out> ByRef CompressedDataSize As Integer) As <MarshalAs(UnmanagedType.Bool)> Boolean
    End Function

    <DllImport("Cabinet.dll", EntryPoint:="Decompress")>
    Private Shared Function Decompress(<[In]> CompressorHandle As IntPtr,
                                       <[In], MarshalAs(UnmanagedType.LPArray)> CompressedData As Byte(),
                                       <[In]> CompressedDataSize As Integer,
                                       <Out, MarshalAs(UnmanagedType.LPArray)> UncompressedBuffer As Byte(),
                                       <[In]> UncompressedBufferSize As Integer,
                                       <Out> ByRef UncompressedDataSize As Integer) As <MarshalAs(UnmanagedType.Bool)> Boolean
    End Function
#End Region

#Region "Enum"
    Public Enum COMPRESS_ALGORITHM As Integer
        MSZIP = COMPRESS_ALGORITHM_MSZIP
        XPRESS = COMPRESS_ALGORITHM_XPRESS
        XPRESS_HUFF = COMPRESS_ALGORITHM_XPRESS_HUFF
        LZMS = COMPRESS_ALGORITHM_LZMS
    End Enum
#End Region

#Region "Const"
    Private Const MagicNumber As Integer = 123456

    Private Const COMPRESS_ALGORITHM_MSZIP As Integer = 2
    Private Const COMPRESS_ALGORITHM_XPRESS As Integer = 3
    Private Const COMPRESS_ALGORITHM_XPRESS_HUFF As Integer = 4
    Private Const COMPRESS_ALGORITHM_LZMS As Integer = 5
#End Region

#Region "Structure"
    Private Structure CompressedDataHeader
        Dim MagicNumber As Integer
        Dim Algorithm As COMPRESS_ALGORITHM
        Dim UncompressedDataSize As Integer
        Sub New(MagicNumber As Integer,
                Algorithm As COMPRESS_ALGORITHM,
                UncompressedDataSize As Integer)
            Me.MagicNumber = MagicNumber
            Me.Algorithm = Algorithm
            Me.UncompressedDataSize = UncompressedDataSize
        End Sub
    End Structure
#End Region

#Region "Functions"
    Public Shared Function CompressData(UncompressedData As Byte(),
                                        ByRef CompressedData As Byte(),
                                        Optional Algorithm As COMPRESS_ALGORITHM = COMPRESS_ALGORITHM.MSZIP) As Boolean
        Dim bolRet As Boolean = False
        If IsAllocated(UncompressedData) AndAlso UncompressedData.Length < Integer.MaxValue \ 2 Then
            Dim CompressorHandle As IntPtr
            If CreateCompressor(Algorithm, IntPtr.Zero, CompressorHandle) Then
                Dim CompressedDataSize As Integer
                If Not Compress(CompressorHandle, UncompressedData, UncompressedData.Length,
                                Nothing, 0, CompressedDataSize) Then
                    Dim CBytes As Byte() = New Byte(CompressedDataSize - 1) {}
                    If Compress(CompressorHandle, UncompressedData, UncompressedData.Length,
                                CBytes, CBytes.Length, CompressedDataSize) Then
                        Array.Resize(CBytes, CompressedDataSize)
                        Dim HBytes As Byte() = StructToByteArray(New CompressedDataHeader(MagicNumber, Algorithm,
                                                                                          UncompressedData.Length))
                        Dim DataOut As Byte() = New Byte(HBytes.Length + CBytes.Length - 1) {}
                        Array.Copy(HBytes, 0, DataOut, 0, HBytes.Length)
                        Array.Copy(CBytes, 0, DataOut, HBytes.Length, CBytes.Length)
                        CompressedData = DataOut
                        bolRet = True
                    End If
                End If
                CloseCompressor(CompressorHandle)
            End If
        End If
        Return bolRet
    End Function

    Public Shared Function UncompressData(CompressedData As Byte(),
                                          ByRef UncompressedData As Byte()) As Boolean
        Dim bolRet As Boolean = False
        If IsAllocated(CompressedData) AndAlso CompressedData.Length > Marshal.SizeOf(Of CompressedDataHeader) Then
            Dim HBytes As Byte() = New Byte(Marshal.SizeOf(Of CompressedDataHeader) - 1) {}
            Array.Copy(CompressedData, 0, HBytes, 0, HBytes.Length)
            Dim CDH As CompressedDataHeader = ByteArrayToStruct(HBytes)
            If CDH.MagicNumber = MagicNumber Then
                Dim CBytes As Byte() = New Byte(CompressedData.Length - HBytes.Length - 1) {}
                Array.Copy(CompressedData, HBytes.Length, CBytes, 0, CBytes.Length)
                Dim DecompressorHandle As IntPtr
                If CreateDecompressor(CDH.Algorithm, IntPtr.Zero, DecompressorHandle) Then
                    Dim DecompressedDataSize As Integer
                    If Not Decompress(DecompressorHandle, CBytes, CBytes.Length,
                                      Nothing, 0, DecompressedDataSize) Then
                        Dim TmpData As Byte() = New Byte(DecompressedDataSize - 1) {}
                        If Decompress(DecompressorHandle, CBytes, CBytes.Length,
                                      TmpData, TmpData.Length, DecompressedDataSize) Then
                            If DecompressedDataSize = CDH.UncompressedDataSize Then
                                Array.Resize(TmpData, DecompressedDataSize)
                                UncompressedData = TmpData
                                bolRet = True
                            End If
                        End If
                    End If
                    CloseDecompressor(DecompressorHandle)
                End If
            End If
        End If
        Return bolRet
    End Function

    Private Shared Function IsAllocated(InArray As Array) As Boolean
        Return Not ((InArray Is Nothing) OrElse (InArray.Length = 0) OrElse (InArray.GetValue(0) Is Nothing))
    End Function

    Private Shared Function StructToByteArray(CDH As CompressedDataHeader) As Byte()
        Dim pCDH As IntPtr = Marshal.AllocHGlobal(Marshal.SizeOf(CDH))
        Dim HBytes As Byte() = New Byte(Marshal.SizeOf(CDH) - 1) {}
        Marshal.StructureToPtr(CDH, pCDH, False)
        Marshal.Copy(pCDH, HBytes, 0, Marshal.SizeOf(CDH))
        Marshal.FreeHGlobal(pCDH)
        Return HBytes
    End Function

    Private Shared Function ByteArrayToStruct(ByteArray As Byte()) As CompressedDataHeader
        Dim pHBytes As IntPtr = Marshal.AllocHGlobal(ByteArray.Length)
        Marshal.Copy(ByteArray, 0, pHBytes, ByteArray.Length)
        Dim CDH As CompressedDataHeader = CType(Marshal.PtrToStructure(pHBytes, GetType(CompressedDataHeader)), CompressedDataHeader)
        Marshal.FreeHGlobal(pHBytes)
        Return CDH
    End Function
#End Region

End Class
